#include "floppy.h"
#include "main.h"

#define FLOPPOTRON_TICK_SPEED (uint16_t)(72000000 / 3600)
#define FLOPPOTRON_SOUND_FREQ (FLOPPOTRON_TICK_SPEED / 2)

#define FLOPPOTRON_MAX_STEPS (100 * 2 * 2)

// класс, описывающий программный таймер для флоппи привода, но мы сделаем вместо него структуру
struct FloppyDrive {
public:
    uint16_t cnt = 0; // текущий счетчик таймера
    uint16_t top = 0; // верхнее значение таймера
    uint16_t curr_step = 0; // текущий шаг флоппи привода
    bool isStarted = false; // запущен ли таймер

    GPIO_TypeDef* gpio{};
    const uint16_t gpioDirBit{}; // бит ноги GPIO для подключения к Floppy::DIR
    const uint16_t gpioStepBit{}; // бит ноги GPIO для подключения к Floppy::STEP
    const uint16_t gpioSelBit{}; // бит ноги GPIO для подключения к Floppy::DRVSB (Drive select B)

    void setFreq(uint16_t freq) {
        // freq = FLOPPOTRON_SOUND_FREQ / top
        top = FLOPPOTRON_SOUND_FREQ / freq;
    }

    void setEnable(bool en) {
        isStarted = en;
    }

    void tick() {
        if (isStarted) {
            cnt++;
            if (cnt >= top) {
                cnt = 0;
                curr_step++;
                if (curr_step >= FLOPPOTRON_MAX_STEPS) {
                    curr_step = 0;
                }
                if (curr_step & 1) {
                    // нечетные шаги, это тик направления
                    if (curr_step < FLOPPOTRON_MAX_STEPS / 2) {
                        gpio->ODR |= gpioDirBit;
                    } else {
                        gpio->ODR &= ~gpioDirBit;
                    }
                } else {
                    // четные шаги, это шаги мотора
                    gpio->ODR ^= gpioStepBit;
                }
            }
        } else {
            cnt = 0;
        }
    }
};

FloppyDrive drives[] = {
        {.gpio = DIR1_GPIO_Port, .gpioDirBit = DIR1_Pin, .gpioStepBit = STEP1_Pin},
        {.gpio = DIR2_GPIO_Port, .gpioDirBit = DIR2_Pin, .gpioStepBit = STEP2_Pin},
};


void FloppySoundTickHandler() {
    for (auto& drive:drives) {
        drive.tick();
    }
}

#define FT_CMD_NOTES_ON 0x1000
#define FT_CMD_NOTES_OFF 0x2000
#define FT_CMD_WAIT 0x3000

// составная команда - частота на
#define FT_CMD_SET_FRQ1 0x8000
#define FT_CMD_SET_FRQ2 0x9000

#define FT_CMD_STOP 0x0000
#define FT_CMD_MASK 0xF000


// хандлер для проигрывания нот
void FloppyTrackTick() {
    // байткод для 100bpm, тестовая мелодия
    const static uint16_t bytecode[] = {
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 165, FT_CMD_WAIT | 50, // '3'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 196, FT_CMD_WAIT | 50, // '5'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 165, FT_CMD_WAIT | 50, // '3'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 196, FT_CMD_WAIT | 50, // '5'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 147, FT_CMD_WAIT | 50, // '2'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 196, FT_CMD_WAIT | 50, // '5'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 147, FT_CMD_WAIT | 50, // '2'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 196, FT_CMD_WAIT | 50, // '5'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 147, FT_CMD_WAIT | 50, // '2'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 174, FT_CMD_WAIT | 50, // '4'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 147, FT_CMD_WAIT | 50, // '2'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 174, FT_CMD_WAIT | 50, // '4'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 147, FT_CMD_WAIT | 50, // '2'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 220, FT_CMD_WAIT | 50, // '6'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 174, FT_CMD_WAIT | 50, // '4'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 220, FT_CMD_WAIT | 50, // '6'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 147, FT_CMD_WAIT | 50, // '2'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 220, FT_CMD_WAIT | 50, // '6'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 174, FT_CMD_WAIT | 50, // '4'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 220, FT_CMD_WAIT | 50, // '6'
//            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 50, // ' '
//            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 50, // ' '
//            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 50, // ' '
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 165, FT_CMD_WAIT | 50, // '3'
//            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 50, // ' '
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 196, FT_CMD_WAIT | 50, // '5'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 174, FT_CMD_WAIT | 50, // '4'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 196, FT_CMD_WAIT | 50, // '5'
//            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 50, // ' '
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 262, FT_CMD_WAIT | 50, // '1''
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 50, // ' '
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 147, FT_CMD_WAIT | 50, // '2'
//            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 50, // ' '
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 174, FT_CMD_WAIT | 50, // '4'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 196, FT_CMD_WAIT | 50, // '5'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 174, FT_CMD_WAIT | 50, // '4'
//            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 50, // ' '
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 196, FT_CMD_WAIT | 50, // '5'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 220, FT_CMD_WAIT | 50, // '6'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 196, FT_CMD_WAIT | 50, // '5'
//            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 50, // ' '
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 147, FT_CMD_WAIT | 50, // '2'
//            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 50, // ' '
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 174, FT_CMD_WAIT | 50, // '4'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 165, FT_CMD_WAIT | 50, // '3'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 174, FT_CMD_WAIT | 50, // '4'
//            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 50, // ' '
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 262, FT_CMD_WAIT | 50, // '1''
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 247, FT_CMD_WAIT | 50, // '7'
//            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 50, // ' '
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 147, FT_CMD_WAIT | 50, // '2'
//            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 50, // ' '
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 174, FT_CMD_WAIT | 50, // '4'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 165, FT_CMD_WAIT | 50, // '3'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 174, FT_CMD_WAIT | 50, // '4'
//            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 50, // ' '
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 196, FT_CMD_WAIT | 50, // '5'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 220, FT_CMD_WAIT | 50, // '6'
//            FT_CMD_NOTES_ON | 1, FT_CMD_SET_FRQ1 | 196, FT_CMD_WAIT | 50, // '5'

            // недо-имперский марш
            FT_CMD_NOTES_ON | 3, FT_CMD_SET_FRQ1 | 165, FT_CMD_SET_FRQ2 | 165, FT_CMD_WAIT | 25, // '3'
            FT_CMD_WAIT | 25, // '_'
            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 25, // ' '
            FT_CMD_NOTES_ON | 3, FT_CMD_SET_FRQ1 | 165, FT_CMD_SET_FRQ2 | 165, FT_CMD_WAIT | 25, // '3'
            FT_CMD_WAIT | 25, // '_'
            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 25, // ' '
            FT_CMD_NOTES_ON | 3, FT_CMD_SET_FRQ1 | 165, FT_CMD_SET_FRQ2 | 165, FT_CMD_WAIT | 25, // '3'
            FT_CMD_WAIT | 25, // '_'
            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 25, // ' '
            FT_CMD_NOTES_ON | 3, FT_CMD_SET_FRQ1 | 131, FT_CMD_SET_FRQ2 | 131, FT_CMD_WAIT | 25, // '1'
            FT_CMD_WAIT | 25, // '_'
            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 25, // ' '
            FT_CMD_NOTES_ON | 3, FT_CMD_SET_FRQ1 | 196, FT_CMD_SET_FRQ2 | 196, FT_CMD_WAIT | 25, // '5'
            FT_CMD_NOTES_ON | 3, FT_CMD_SET_FRQ1 | 165, FT_CMD_SET_FRQ2 | 165, FT_CMD_WAIT | 25, // '3'
            FT_CMD_WAIT | 25, // '_'
            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 25, // ' '
            FT_CMD_NOTES_ON | 3, FT_CMD_SET_FRQ1 | 131, FT_CMD_SET_FRQ2 | 131, FT_CMD_WAIT | 25, // '1'
            FT_CMD_WAIT | 25, // '_'
            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 25, // ' '
            FT_CMD_NOTES_ON | 3, FT_CMD_SET_FRQ1 | 196, FT_CMD_SET_FRQ2 | 196, FT_CMD_WAIT | 25, // '5'
            FT_CMD_NOTES_ON | 3, FT_CMD_SET_FRQ1 | 165, FT_CMD_SET_FRQ2 | 165, FT_CMD_WAIT | 25, // '3'
            FT_CMD_WAIT | 25, // '_'

            FT_CMD_NOTES_OFF | 3, FT_CMD_WAIT | 500,
            FT_CMD_STOP
    };

    static const uint16_t* pc = bytecode;
    static uint16_t wait = 0;

    if (wait == 0) {
        while (wait == 0) {
            uint16_t cmd = *(pc++);
            uint16_t value = cmd & (~FT_CMD_MASK);
            switch (cmd & FT_CMD_MASK) {
                case FT_CMD_STOP:
                    pc = bytecode; // просто все начнем сначала
                    drives[0].setEnable(false);
                    drives[1].setEnable(false);
                    break;

                case FT_CMD_NOTES_ON:
                    if (value & 1) {
                        drives[0].setEnable(true);
                    }
                    if (value & 2) {
                        drives[1].setEnable(true);
                    }
                    break;

                case FT_CMD_NOTES_OFF:
                    if (value & 1) {
                        drives[0].setEnable(false);
                    }
                    if (value & 2) {
                        drives[1].setEnable(false);
                    }
                    break;

                case FT_CMD_WAIT:
                    wait = value;
                    break;

                case FT_CMD_SET_FRQ1:
                    drives[0].setFreq(value);
                    break;

                case FT_CMD_SET_FRQ2:
                    drives[1].setFreq(value);
                    break;
            }
        }
    } else {
        wait--;
    }
}